@@ -1,20 +1,35 @@
+import json
+import os
+import pprint
+import re
+from collections import defaultdict
 from concurrent.futures import ThreadPoolExecutor
-from pathlib import Path
-from shutil import copyfileobj
-from typing import Optional, List
-
-import pandas as pd
-import numpy as np
-import requests, json, os, pprint
 from copy import deepcopy
 from io import BytesIO
+from pathlib import Path
+from shutil import copyfileobj
+from typing import Optional, List, Iterable, Dict, Union
+
+import numpy as np
+import pandas as pd
+import requests
 from PIL import Image
 
+from metaspace.image_processing import clip_hotspots
+
 try:
     from typing import TypedDict  # Requires Python 3.8
 except ImportError:
     TypedDict = dict
 
+try:
+    from pandas import json_normalize  # Only in Pandas 1.0.0+
+except ImportError:
+    from pandas.io.json import json_normalize  # Logs DeprecationWarning if used in Pandas 1.0.0+
+
+
+DEFAULT_DATABASE = 'HMDB-v4'
+
 
 class DatasetDownloadLicense(TypedDict):
     code: str
@@ -38,21 +53,50 @@ class DatasetDownload(TypedDict):
     files: List[DatasetDownloadFile]
 
 
+class MetaspaceException(Exception):
+    pass
+
+
+class GraphQLException(MetaspaceException):
+    def __init__(self, json, message):
+        super().__init__(message)
+        self.json = json
+        self.message = message
+
+
+class BadRequestException(MetaspaceException):
+    def __init__(self, json, message, type=None):
+        super().__init__(f"{type}: {message}")
+        self.json = json
+        self.message = message
+        self.type = type
+
+
+class InvalidResponseException(MetaspaceException):
+    def __init__(self, json, http_response):
+        super().__init__('Invalid response from server')
+        self.json = json
+        self.http_response = http_response
+
+
 def _extract_data(res):
     if not res.headers.get('Content-Type').startswith('application/json'):
-        raise Exception('Wrong Content-Type: {}'.format(res.headers.get('Content-Type')))
+        raise Exception(
+            'Wrong Content-Type: {}.\n{}'.format(res.headers.get('Content-Type'), res.text)
+        )
+
     res_json = res.json()
     if 'data' in res_json and 'errors' not in res_json:
         return res_json['data']
     else:
         if 'errors' in res_json:
             pprint.pprint(res_json['errors'])
-            raise Exception(res_json['errors'][0]['message'])
+            raise GraphQLException(res_json, res_json['errors'][0]['message'])
         elif 'message' in res_json:
-            raise Exception(res_json['message'])
+            raise BadRequestException(res_json, res_json['message'], res_json.get('type'))
         else:
             pprint.pprint(res_json)
-            raise Exception('Invalid response from server')
+            raise InvalidResponseException(res_json, res)
 
 
 def get_config(
@@ -213,7 +257,7 @@ class GraphQLClient(object):
         configJson
         metadataJson
         isPublic
-        molDBs
+        databases { id name version isPublic archived }
         adducts
         acquisitionGeometry
         metadataType
@@ -238,13 +282,9 @@ class GraphQLClient(object):
         offSampleProb
         dataset { id name }
         possibleCompounds { name information { url databaseId } }
-        isotopeImages { mz url maxIntensity totalIntensity }
+        isotopeImages { mz url minIntensity maxIntensity totalIntensity }
     """
 
-    DEFAULT_ANNOTATION_FILTER = {
-        'database': 'HMDB-v4',
-    }
-
     def getDataset(self, datasetId):
         query = (
             """
@@ -337,17 +377,10 @@ class GraphQLClient(object):
             ','.join(query_arguments),
             self.ANNOTATION_FIELDS,
         )
-        if datasetFilter is None:
-            datasetFilter = {}
-
-        if annotationFilter is None:
-            annot_filter = {}
-        else:
-            annot_filter = deepcopy(annotationFilter)
-            for key, val in self.DEFAULT_ANNOTATION_FILTER.items():
-                annot_filter.setdefault(key, val)
+        annot_filter = annotationFilter
 
         if colocFilter:
+            annot_filter = deepcopy(annot_filter) if annot_filter else {}
             annot_filter.update(colocFilter)
             order_by = 'ORDER_BY_COLOCALIZATION'
         else:
@@ -370,18 +403,12 @@ class GraphQLClient(object):
                 datasetFilter: $dFilter
               )
             }"""
-        if datasetFilter is None:
-            annotationFilter = {}
-        if annotationFilter is None:
-            annotFilter = {}
-        else:
-            annotFilter = deepcopy(annotationFilter)
-            for key, val in self.DEFAULT_ANNOTATION_FILTER.items():
-                annotFilter.setdefault(key, val)
 
-        return self.query(query=query, variables={'filter': annotFilter, 'dFilter': datasetFilter})
+        return self.query(
+            query=query, variables={'filter': annotationFilter, 'dFilter': datasetFilter}
+        )
 
-    def getDatasets(self, datasetFilter={}):
+    def getDatasets(self, datasetFilter=None):
         query = (
             """
         query getDatasets($filter: DatasetFilter,
@@ -420,14 +447,49 @@ class GraphQLClient(object):
         variables = {"datasetId": dsid}
         return self.query(query, variables)
 
+    def get_databases(self):
+        query = """
+            {
+              molecularDatabases {
+                id name version isPublic archived
+              }
+            }
+        """
+        result = self.query(query)
+        return result['molecularDatabases']
+
+    def map_database_to_id(self, database):
+        # Forwards/backwards compatibility issue: the GraphQL Schema may soon change from Int ids
+        # to ID (i.e. str-based) ids. For now, this supports both types, and passes the type on
+        # without modification. When the API has settled, this should be updated to coerce to the
+        # correct type, because we shouldn't burden users with having to figure out why calls are
+        # failing when they pass IDs that look like integers as integers instead of strings.
+        if isinstance(database, int) or re.match('^\d+$', database):
+            return database
+
+        database_docs = self.get_databases()
+        database_name_id_map = defaultdict(list)
+        for db in database_docs:
+            database_name_id_map[db['name']].append(db['id'])
+
+        database_ids = database_name_id_map.get(database, [])
+        if len(database_ids) == 0:
+            raise Exception(
+                f'Database not found or you do not have access to it. Available databases: '
+                f'{list(database_name_id_map.keys())}'
+            )
+        if len(database_ids) > 1:
+            raise Exception(f'Database name "{database}" is not unique. Use database id instead.')
+
+        return database_ids[0]
+
     def create_dataset(
         self,
         data_path,
         metadata,
-        priority=0,
         ds_name=None,
         is_public=None,
-        mol_dbs=None,
+        databases=None,
         adducts=None,
         ppm=None,
         ds_id=None,
@@ -448,7 +510,7 @@ class GraphQLClient(object):
                 'name': ds_name,
                 'inputPath': data_path,
                 'isPublic': is_public,
-                'molDBs': mol_dbs,
+                'databaseIds': databases and [self.map_database_to_id(db) for db in databases],
                 'adducts': adducts,
                 'ppm': ppm,
                 'submitterId': submitter_id,
@@ -473,7 +535,7 @@ class GraphQLClient(object):
         self,
         ds_id,
         name=None,
-        mol_dbs=None,
+        databases=None,
         adducts=None,
         ppm=None,
         reprocess=False,
@@ -495,8 +557,10 @@ class GraphQLClient(object):
         input_field = {}
         if name:
             input_field['name'] = name
-        if mol_dbs:
-            input_field['molDBs'] = mol_dbs
+
+        if databases:
+            input_field['databaseIds'] = [self.map_database_to_id(db) for db in databases]
+
         if adducts:
             input_field['adducts'] = adducts
         if ppm:
@@ -511,21 +575,6 @@ class GraphQLClient(object):
 
         return self.query(query, variables)
 
-    def get_molecular_databases(self, names=None):
-        query = '''{
-          molecularDatabases {
-            id name version
-          }
-        }'''
-        resp = self.query(query)
-        if 'molecularDatabases' not in resp:
-            raise Exception('GraphQL error: {}'.format(resp))
-        mol_dbs = resp['molecularDatabases']
-        if not names:
-            return mol_dbs
-        else:
-            return [d for d in mol_dbs if d['name'] in names]
-
 
 def ion(r):
     from pyMSpec.pyisocalc.tools import normalise_sf
@@ -636,7 +685,7 @@ NYI = Exception("NOT IMPLEMENTED YET")
 class SMDataset(object):
     def __init__(self, _info, gqclient):
         self._info = _info
-        self._gqclient = gqclient
+        self._gqclient: GraphQLClient = gqclient
         self._config = json.loads(self._info['configJson'])
         self._metadata = Metadata(self._info['metadataJson'])
         self._session = requests.session()
@@ -657,11 +706,25 @@ class SMDataset(object):
         return "SMDataset({} | ID: {})".format(self.name, self.id)
 
     def annotations(
-        self, fdr=0.1, database=None, return_vals=('sumFormula', 'adduct'), **annotation_filter
-    ):
+        self,
+        fdr: float = 0.1,
+        database: Union[str, int] = DEFAULT_DATABASE,
+        return_vals: Iterable = ('sumFormula', 'adduct'),
+        **annotation_filter,
+    ) -> List[list]:
+        """Fetch dataset annotations.
+
+        Args:
+            fdr: Max FDR level.
+            database: Database name or id.
+            return_vals: Tuple of fields to return.
+
+        Returns:
+            List of annotations with requested fields.
+        """
         annotation_filter.setdefault('fdrLevel', fdr)
         if database:
-            annotation_filter['database'] = database
+            annotation_filter['databaseId'] = self._gqclient.map_database_to_id(database)
         dataset_filter = {'ids': self.id}
 
         records = self._gqclient.getAnnotations(annotation_filter, dataset_filter)
@@ -669,13 +732,26 @@ class SMDataset(object):
 
     def results(
         self,
-        database,
-        fdr=None,
-        coloc_with=None,
-        include_chem_mods=False,
-        include_neutral_losses=False,
+        database: Union[str, int] = DEFAULT_DATABASE,
+        fdr: float = None,
+        coloc_with: str = None,
+        include_chem_mods: bool = False,
+        include_neutral_losses: bool = False,
         **annotation_filter,
-    ):
+    ) -> pd.DataFrame:
+        """Fetch all dataset annotations as dataframe.
+
+        Args:
+            database: Molecular database name or id.
+            fdr: Max FDR level.
+            coloc_with: Fetch only results colocalized with formula.
+            include_chem_mods: Include results with chemical modifications.
+            include_neutral_losses: Include results with neutral losses.
+
+        Returns:
+            List of annotations with requested fields.
+        """
+
         def get_compound_database_ids(possibleCompounds):
             return [
                 compound['information'] and compound['information'][0]['databaseId']
@@ -685,14 +761,16 @@ class SMDataset(object):
         if coloc_with:
             assert fdr
             coloc_coeff_filter = {
-                'database': database,
                 'colocalizedWith': coloc_with,
                 'fdrLevel': fdr,
             }
+            if database:
+                coloc_coeff_filter['databaseId'] = self._gqclient.map_database_to_id(database)
             annotation_filter.update(coloc_coeff_filter)
         else:
             coloc_coeff_filter = None
-            annotation_filter['database'] = database
+            if database:
+                annotation_filter['databaseId'] = self._gqclient.map_database_to_id(database)
             if fdr:
                 annotation_filter['fdrLevel'] = fdr
 
@@ -714,7 +792,7 @@ class SMDataset(object):
         if not records:
             return pd.DataFrame()
 
-        df = pd.io.json.json_normalize(records)
+        df = json_normalize(records)
         return (
             df.assign(
                 moleculeNames=df.possibleCompounds.apply(
@@ -754,12 +832,18 @@ class SMDataset(object):
     def polarity(self):
         return 'Positive' if self._config['isotope_generation']['charge'] > 0 else 'Negative'
 
+    @property
+    def database_details(self):
+        return self._info['databases']
+
     @property
     def databases(self):
-        return self._config['databases']
+        """DEPRECATED. Use 'databases_details' instead."""
+        return [d['name'] for d in self.database_details]
 
     @property
     def database(self):
+        """DEPRECATED. Use 'databases_details' instead."""
         return self.databases[0]
 
     @property
@@ -776,6 +860,7 @@ class SMDataset(object):
         adduct,
         only_first_isotope=False,
         scale_intensity=True,
+        hotspot_clipping=False,
         neutral_loss='',
         chem_mod='',
     ):
@@ -788,6 +873,8 @@ class SMDataset(object):
                                          are usually lower quality copies of the first isotopic ion image.
         :param bool  scale_intensity:    When True, the output values will be scaled to the intensity range of the original data.
                                          When False, the output values will be in the 0.0 to 1.0 range.
+        :param bool  hotspot_clipping:   When True, apply hotspot clipping. Recommended if the images will be used for visualisation.
+                                         This is required to get ion images that match the METASPACE website
         :param str   neutral_loss:
         :param str   chem_mod:
         :return IsotopeImages:
@@ -809,7 +896,10 @@ class SMDataset(object):
             if not url:
                 return None
             url = self._baseurl + url
-            im = mpimg.imread(BytesIO(self._session.get(url).content))
+            try:
+                im = mpimg.imread(BytesIO(self._session.get(url).content))
+            except:
+                im = mpimg.imread(BytesIO(self._session.get(url).content))
             mask = im[:, :, 3]
             data = im[:, :, 0]
             data[mask == 0] = 0
@@ -834,29 +924,45 @@ class SMDataset(object):
                 if images[i] is None:
                     images[i] = np.zeros(shape, dtype=non_empty_images[0].dtype)
                 else:
-                    images[i] *= float(image_metadata[i]['maxIntensity'])
+                    lo = float(image_metadata[i]['minIntensity'])
+                    hi = float(image_metadata[i]['maxIntensity'])
+                    images[i] = lo + images[i] * (hi - lo)
+
+        if hotspot_clipping:
+            for i in range(len(images)):
+                if images[i] is not None:
+                    images[i] = clip_hotspots(images[i])
+                    if not scale_intensity:
+                        # Renormalize to 0-1 range
+                        images[i] /= np.max(images[i]) or 1
 
         return IsotopeImages(images, sf, chem_mod, neutral_loss, adduct, image_mzs, image_urls)
 
     def all_annotation_images(
         self,
-        fdr=0.1,
-        database=None,
-        only_first_isotope=False,
-        scale_intensity=True,
+        fdr: float = 0.1,
+        database: Union[str, int] = DEFAULT_DATABASE,
+        only_first_isotope: bool = False,
+        scale_intensity: bool = True,
+        hotspot_clipping: bool = False,
         **annotation_filter,
-    ):
-        """
-        Retrieve all ion images for the dataset and given annotation filters.
-        :param float fdr:                Maximum FDR level of annotations
-        :param str   database:           Molecular database to use (e.g. HMDBv4)
-        :param bool  only_first_isotope: Only retrieve the first (most abundant) isotopic ion image for each annotation.
-                                         Typically this is all you need for data analysis, as the less abundant isotopes
-                                         are usually lower quality copies of the first isotopic ion image.
-        :param bool  scale_intensity:    When True, the output values will be scaled to the intensity range of the original data.
-                                         When False, the output values will be in the 0.0 to 1.0 range.
-        :param       annotation_filter:  Additional filters passed to `SMDataset.annotations`
-        :return list[IsotopeImages]:
+    ) -> List[IsotopeImages]:
+        """Retrieve all ion images for the dataset and given annotation filters.
+
+        Args:
+            fdr: Maximum FDR level of annotations.
+            database: Molecular database name or id.
+            only_first_isotope: Only retrieve the first (most abundant) isotopic ion image for each
+                annotation. Typically this is all you need for data analysis, as the less abundant
+                isotopes are usually lower quality copies of the first isotopic ion image.
+            scale_intensity: When True, the output values will be scaled to the intensity range of
+                the original data. When False, the output values will be in the 0.0 to 1.0 range.
+            hotspot_clipping:   When True, apply hotspot clipping. Recommended if the images will
+                be used for visualisation. This is required to get ion images that match the
+                METASPACE website
+            annotation_filter: Additional filters passed to `SMDataset.annotations`.
+        Returns:
+            list of isotope images
         """
         with ThreadPoolExecutor() as pool:
 
@@ -869,6 +975,7 @@ class SMDataset(object):
                     scale_intensity=scale_intensity,
                     neutral_loss=neutral_loss,
                     chem_mod=chem_mod,
+                    hotspot_clipping=hotspot_clipping,
                 )
 
             annotations = self.annotations(
@@ -938,8 +1045,26 @@ class SMDataset(object):
 
 
 class SMInstance(object):
-    def __init__(self, *args, **kwargs):
-        self._config = get_config(*args, **kwargs)
+    def __init__(
+        self,
+        host: str = None,
+        verify_certificate: bool = True,
+        email: str = None,
+        password: str = None,
+        api_key: str = None,
+        config_path: str = None,
+    ):
+        """Client class for communication with the Metaspace API.
+
+        Args:
+            host: Full host name, e.g. 'https://metaspace2020.eu'
+            verify_certificate: Ignore certificate validation.
+            email: User email.
+            password: User password.
+            api_key: User API key.
+            config_path: Configuration file path.
+        """
+        self._config = get_config(host, verify_certificate, email, password, api_key, config_path)
         if not self._config['verify_certificate']:
             import warnings
             from urllib3.exceptions import InsecureRequestWarning
@@ -995,20 +1120,21 @@ class SMInstance(object):
     def all_adducts(self):
         raise NYI
 
-    def database(self, name, version=None):
-        assert self._moldb_client, 'provide moldb_url in the config'
-        return self._moldb_client.getDatabase(name, version)
+    def database(self, name: str = None, version: str = None, id: int = None) -> Dict:
+        """Fetch molecular database by id."""
+        databases = self._gqclient.get_databases()
+        results = [db for db in databases if db['id'] == id or db['name'] == name]
+        return results[0] if len(results) > 0 else None
 
     def databases(self):
-        assert self._moldb_client, 'provide moldb_url in the config'
-        return self._moldb_client.getDatabaseList()
+        return self._gqclient.get_databases()
 
     def metadata(self, datasets):
         """
         Pandas dataframe for a subset of datasets
         where rows are flattened metadata JSON objects
         """
-        df = pd.io.json.json_normalize([d.metadata.json for d in datasets])
+        df = json_normalize([d.metadata.json for d in datasets])
         df.index = [d.name for d in datasets]
         return df
 
@@ -1021,7 +1147,7 @@ class SMInstance(object):
         records = self._gqclient.getAnnotations(
             annotationFilter={'database': db_name, 'fdrLevel': fdr}, datasetFilter=datasetFilter
         )
-        df = pd.io.json.json_normalize(records)
+        df = json_normalize(records)
         return pd.DataFrame(
             dict(
                 formula=df['sumFormula'],
@@ -1042,13 +1168,12 @@ class SMInstance(object):
         datasets = self._gqclient.getDatasets(datasetFilter=datasetFilter)
         df = pd.concat(
             [
-                pd.DataFrame(pd.io.json.json_normalize(json.loads(dataset['metadataJson'])))
+                pd.DataFrame(json_normalize(json.loads(dataset['metadataJson'])))
                 for dataset in datasets
             ]
         )
         df.index = [dataset['id'] for dataset in datasets]
         return df
-        # return pd.DataFrame.from_records([pd.io.json.json_normalize(json.loads(dataset['metadataJson'])) for dataset in datasets])
 
     def submit_dataset(
         self,
@@ -1064,8 +1189,7 @@ class SMInstance(object):
         s3bucket=None,
         priority=0,
     ):
-        """
-        Submit a dataset for processing on the SM Instance
+        """DEPRECATED. Submit a dataset for processing on the SM Instance
         :param imzml_fn: file path to imzml
         :param ibd_fn: file path to ibd
         :param metadata: a properly formatted metadata json string
@@ -1095,39 +1219,40 @@ class SMInstance(object):
         return self._gqclient.create_dataset(
             data_path=folder,
             metadata=metadata,
-            priority=priority,
             ds_name=dsName,
             is_public=isPublic,
-            mol_dbs=molDBs,
+            databases=molDBs,
             adducts=adducts,
             ds_id=dsid,
         )
 
     def submit_dataset_v2(
         self,
-        imzml_fn,
-        ibd_fn,
-        ds_name,
-        metadata,
+        imzml_fn: str,
+        ibd_fn: str,
+        ds_name: str,
+        metadata: str,
         s3_bucket,
-        is_public=None,
-        moldbs=None,
-        adducts=None,
-        priority=0,
+        is_public: bool = None,
+        moldbs: Union[List[str], List[int]] = None,
+        adducts: List[str] = None,
+        priority: int = 0,
     ):
-        """
-        Submit a dataset for processing on the SM Instance
+        """Submit a dataset for processing in Metaspace.
 
-        :param imzml_fn: file path to imzml
-        :param ibd_fn: file path to ibd
-        :param ds_name: dataset name
-        :param metadata: a properly formatted metadata json string
-        :param s3_bucket: boto3 s3 bucket object, both the user has write permission to
-        and METASPACE can access
-        :param is_public: make dataset public
-        :param moldbs: list molecular databases
-        :param adducts: list of adducts
-        :return:
+        Args:
+            imzml_fn: Imzml file path.
+            ibd_fn: Ibd file path.
+            ds_name: Dataset name.
+            metadata: Properly formatted metadata json string.
+            s3_bucket: boto3 s3 bucket object, both the user has write permission to
+                and METASPACE can access.
+            is_public: Make dataset public.
+            moldbs: List of databases.
+            adducts: List of adducts.
+
+        Returns:
+            Request status.
         """
         import uuid
 
@@ -1138,16 +1263,15 @@ class SMInstance(object):
         return self._gqclient.create_dataset(
             data_path=f's3a://{s3_bucket.name}/{dataset_key}',
             metadata=metadata,
-            priority=priority,
             ds_name=ds_name,
             is_public=is_public,
-            mol_dbs=moldbs,
+            databases=moldbs,
             adducts=adducts,
         )
 
     def update_dataset_dbs(self, datasetID, molDBs=None, adducts=None, priority=1):
         return self._gqclient.update_dataset(
-            ds_id=datasetID, mol_dbs=molDBs, adducts=adducts, priority=priority, reprocess=True
+            ds_id=datasetID, databases=molDBs, adducts=adducts, reprocess=True, priority=priority
         )
 
     def reprocess_dataset(self, dataset_id, force=False):
@@ -1215,6 +1339,8 @@ class SMInstance(object):
 
 
 class MolecularDatabase:
+    """DEPRECATED. Use 'SMInstance.databases' instead."""
+
     def __init__(self, metadata, client):
         self._metadata = metadata
         self._id = self._metadata['id']
